/**
 * Copyright(C) 2013 Patrik Dufresne Service Logiciel <info@patrikdufresne.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.planimod.core.planif;

import org.apache.commons.lang3.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.patrikdufresne.ilp.util.BoundSnapshot;
import com.planimod.core.Team;

/**
 * This runnable is used to check feasibility of assigning employee to preferred team.
 * 
 * @author Patrik Dufresne
 */
public class CheckPreferredTeam extends SolveRunnable {

    private static final transient Logger LOGGER = LoggerFactory.getLogger(CheckPreferredTeam.class);

    /**
     * Bound snapshot used to restore the bound is the problem is unfeasible.
     */
    private BoundSnapshot boundSnapshotXijk;
    /**
     * Bound snapshot used to restore the bound is the problem is unfeasible.
     */
    private BoundSnapshot boundSnapshotXij;

    /**
     * The employee index.
     */
    private final int i;

    /**
     * The team.
     */
    private final Team team;

    /**
     * Create a new runnable to assign employee <code>i</code> to preferred team <code>j</code>
     * 
     * @param problem
     *            reference to ProblemBuilder
     * @param i
     *            the employee
     * @param team
     *            the preferred team
     */
    public CheckPreferredTeam(ProblemBuilder problem, int i, Team team) {
        super(problem);
        this.i = i;
        Validate.notNull(this.team = team);
    }

    @Override
    protected void feasible() {
        LOGGER.info("Employee {} assign to team {}.", //$NON-NLS-1$
                pb.employeeToString(i),
                teamToString(team));
    }

    /**
     * This implementation is using the snapshot to determine if the employee is laready assigned to the team. If so, it
     * also check if the employee if assign only to unswappable task.
     * <p>
     * If the employee is assign to swappable task, this runnable will try to improve the current solution.
     */
    @Override
    protected boolean setUp() {
        // Check if employee can be assigned to team j
        if (pb.xij().iEq(i).jEq(team).isEmpty()) {
            this.feasible = false;
            return false;
        }
        // Take bounds snapshot before changing them

        this.boundSnapshotXijk = pb.xijk().iEq(i).createBoundSnapshot();
        this.boundSnapshotXij = pb.xij().iEq(i).createBoundSnapshot();
        // Fix to zero (0) xik, xij and xi variables not corresponding to
        // the team for employee i. Be careful not to change fixed
        // variables.
        // TODO Check if it really make a different to "fixed" xijk variables.
        pb.xijk().iEq(i).jNotEq(team).notFixed().setBounds(ZERO, ZERO);
        pb.xij().iEq(i).jNotEq(team).notFixed().setBounds(ZERO, ZERO);
        pb.xij().iEq(i).jEq(team).notFixed().setBounds(ONE, ONE);
        // Don't create any constraint to force the employee to work a
        // minimum number of time on the team. OItherwise it will break the
        // non-availability handler. See ticket #191.
        // Sets the objective
        pb.lp.setObjectiveLinear(null);
        LOGGER.debug("Check feasiblility with employee {} " //$NON-NLS-1$
                + "working on team {}.", pb.employeeToString(i), //$NON-NLS-1$
                teamToString(team));
        return true;
    }

    /**
     * Return a string representation of the team j
     * 
     * @param j
     *            the team
     * @return the string
     */
    private String teamToString(Team team) {
        StringBuilder buf = new StringBuilder();
        buf.append(pb.getTeams().indexOf(team));
        String name = team.getName();
        if (name != null) {
            buf.append(" ("); //$NON-NLS-1$
            buf.append(team.getName());
            buf.append(")"); //$NON-NLS-1$
        }
        return buf.toString();
    }
    
    @Override
    protected void tearDown() throws GeneratePlanifException {
        // Nothing to do.
    }

    @Override
    protected void unfeasible() {
        if (this.boundSnapshotXijk != null) {
            // Restore the bound snapshot.
            this.boundSnapshotXijk.restore();
            this.boundSnapshotXijk = null;
            this.boundSnapshotXij.restore();
            this.boundSnapshotXij = null;
            LOGGER.info("Can't assign " //$NON-NLS-1$
                    + "employee {} to team {}", pb.employeeToString(i), //$NON-NLS-1$
                    teamToString(team));
        }
    }
}