/**
 * Copyright(C) 2013 Patrik Dufresne Service Logiciel <info@patrikdufresne.com>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package com.planimod.core.planif;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.patrikdufresne.ilp.Constraint;
import com.patrikdufresne.ilp.Linear;
import com.patrikdufresne.ilp.LinearProblem;
import com.planimod.core.EmployeePreference;
import com.planimod.core.Section;

/**
 * Used to maximize the number of assigned to preferred section.
 * <p>
 * This objective doesn't maximize the preferred section of swappable task. See bug #68.
 * 
 * @author Patrik Dufresne
 */
@SuppressWarnings("unqualified-field-access")
public class MaxPreferredSection extends SolveRunnable {

    private static final transient Logger LOGGER = LoggerFactory.getLogger(MaxPreferredSection.class);

    /**
     * The employee's preferred section or null
     */
    private final Section section;

    /**
     * The employee.
     */
    private final int i;

    /**
     * The linear being maximized.
     */
    private Linear linear;

    /**
     * Default constructor to maximize the preferred section of the employee <code>i</code>.
     * 
     * @param problem
     *            reference to ProblemBuilder
     * 
     * @param i
     *            the employee
     */
    public MaxPreferredSection(ProblemBuilder problem, int i) {
        super(problem);
        this.i = i;
        // Get preferred section for the given employee.
        EmployeePreference pref = pb.preferences.get(i);
        this.section = pref == null ? null : pref.getPreferredSection();
    }

    private void assignEmployeeToSection(Integer lowerBound) {
        // Add a constraint to make sure the employee is assigned to the
        // preferred section.
        String name = String.format("MaxPreferredSection_gte_%d_%d", lowerBound, i); //$NON-NLS-1$
        pb.lp.addConstraint(name, this.linear, lowerBound, null);
    }

    @Override
    protected void feasible() {
        // Assign employee to preferred section.
        Integer value = Integer.valueOf(pb.lp.getObjectiveValue().intValue());
        assignEmployeeToSection(value);
    }

    @Override
    protected boolean setUp() {
        // Check if the employee is schedule to work according to xi bounds
        if (pb.xi().iEq(i).isFixedTo(ZERO)) {
            LOGGER.debug("Employee {} " //$NON-NLS-1$
                    + " is not working according to current values.", //$NON-NLS-1$
                    pb.employeeToString(i));
            this.feasible = false;
            return false;
        }
        // Check if the employee as preferred section.
        // And check if preferred section is part of the planif.
        if (section == null || !pb.getSections().contains(section)) {
            this.feasible = false;
            return false;
        }
        // Build the linear expression to be maximize
        this.linear = pb.xijk().iEq(i).kInRTj().kInSection(section).createLinear();
        if (this.linear.isEmpty()) {
            this.feasible = false;
            return false;
        }
        // Set the linear as an objective
        pb.lp.setObjectiveLinear(this.linear);
        pb.lp.setObjectiveDirection(LinearProblem.MAXIMIZE);
        LOGGER.info("Maximize employee {} preferred section.", //$NON-NLS-1$
                pb.employeeToString(i));
        return true;
    }

    @Override
    protected void tearDown() {
        // Remove objective
        pb.lp.setObjectiveLinear(null);
    }

    @Override
    protected void unfeasible() {
        if (pb.getSections().contains(section)) {
            LOGGER.warn("Employee {} can't be assigned to " //$NON-NLS-1$
                    + "preferred section.", pb.employeeToString(i)); //$NON-NLS-1$
        }
    }
}