/**
 * Copyright(C) 2013 Patrik Dufresne Service Logiciel <info@patrikdufresne.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.planimod.core.planif;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.patrikdufresne.ilp.Linear;
import com.patrikdufresne.ilp.LinearProblem;

/**
 * Minimize the number of employee assigned.
 * 
 * @author Patrik Dufresne
 */
@SuppressWarnings("unqualified-field-access")
public class MinEmployeeAssigned extends AbstractConstraintRunnable {

    private static final transient Logger LOGGER = LoggerFactory.getLogger(MinEmployeeAssigned.class);

    /**
     * The constraint's name.
     * <pre>
     * SUM from{i} to{n} xi <= %f
     * </pre>
     */
    private static final String CONSTRAINT_NAME = "MinEmployeeAssigned_%f"; //$NON-NLS-1$

    /**
     * The linear being minimized.
     */
    private Linear linear;

    /**
     * Public constructor.
     * 
     * @param problem
     *            reference to ProblemBuilder
     */
    public MinEmployeeAssigned(ProblemBuilder problem) {
        super(problem);
    }

    @Override
    protected void feasible() {
        // Get the objective value
        Double value = Double.valueOf(pb.lp.getObjectiveValue().doubleValue());
        fixMaximumNumberOfEmployee(value);
    }

    /**
     * Create a redefine the constraint to sets the maximum number of employees.
     * 
     * @param value
     *            the maximum number of employee.
     */
    private void fixMaximumNumberOfEmployee(Double value) {
        // Create or redefine the constraint upper bound.
        String name = String.format(CONSTRAINT_NAME, value);
        if (this.constraint != null) {
            this.constraint.dispose();
        }
        this.constraint = pb.lp.addConstraint(name, linear, null, value);
        LOGGER.info("Minimum number of employees: {}", value); //$NON-NLS-1$
        this.feasible = true;
    }

    @Override
    protected boolean setUp() {
        LOGGER.debug("Minimize number of employee."); //$NON-NLS-1$
        // Create the following linear function : SUM xi
        this.linear = pb.xi().createLinear();
        // Set the linear as an objective.
        pb.lp.setObjectiveLinear(this.linear);
        pb.lp.setObjectiveDirection(LinearProblem.MINIMIZE);
        return true;
    }

    @Override
    protected void tearDown() {
        // Remove objective
        pb.lp.setObjectiveLinear(null);
        this.linear = null;
    }

    @Override
    protected void unfeasible() throws GeneratePlanifException {
        // It's a maximization problem, it' must be feasible.
        throwProblemUnFeasible();
    }
}