/**
 * Copyright(C) 2013 Patrik Dufresne Service Logiciel <info@patrikdufresne.com>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package com.planimod.core.planif;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.Validate;
import org.apache.commons.lang3.mutable.MutableInt;

import com.patrikdufresne.ilp.Linear;
import com.patrikdufresne.ilp.Variable;
import com.patrikdufresne.ilp.util.BoundSnapshot;

/**
 * Gives a simple interface to manipulate the xi variables
 * <p>
 * Method chaining will filter the variables to construct a readable expression. Can also use operations on each
 * variables on selected group of variables.
 * 
 * @param pb
 *            the problemBuilder containing the array of variables to manipulate.
 * @author Karl
 * 
 */
public class XiVariableSet {

    /**
     * Close you execute for each variable.
     * 
     * @author Patrik Dufresne
     * 
     */
    public static interface Closure {
        public void execute(Variable xi, int i);
    }

    /**
     * Triggers when a variable is not Empty.
     * 
     * @see XiVariableSet#isEmpty()
     * @author Karl
     */
    private static class NotEmptyException extends RuntimeException {
        // Nothing to do
    }

    /**
     * Triggers when a variable is not fixed.
     * 
     * @see XiVariableSet#isFixedTo(Double)
     * @author Karl
     */
    private static class NotFixedException extends RuntimeException {
        // Nothing to do
    }

    /**
     * Predicate to trim the variables
     * 
     * @author Patrik Dufresne
     * 
     */
    public static interface Predicate {
        public boolean evaluate(Variable xi, int i);
    }

    /**
     * Trigger when a variable value is not equal to the given <code>value</code>.
     * 
     * @see XiVariableSet#valueEq()
     * @author Karl
     */
    private static class ValueNotEqException extends RuntimeException {
        // Nothing to do
    }

    /**
     * Value of i, the index of the employee
     */
    private Integer iEq;

    /**
     * The problemBuilder associated
     */
    private ProblemBuilder pb;

    /**
     * Predicate list.
     */
    private List<Predicate> predicates;

    /**
     * Gives a simple interface to manipulate the xi variables.
     * <p>
     * Method chaining will filter the variables to construct a readable expression. Can also use operations on each
     * variables on selected group of variables.
     * 
     * @param pb
     *            the problemBuilder containing the array of variables to manipulate.
     */
    protected XiVariableSet(ProblemBuilder pb) {
        Validate.notNull(this.pb = pb);
    }

    /**
     * Closure creating a snapshot of the current xi variables
     * 
     * @return the BoundSnapshot of the xi variables
     */
    public BoundSnapshot createBoundSnapshot() {
        final List<Variable> list = new ArrayList<Variable>();
        execute(new Closure() {
            @Override
            public void execute(Variable xi, int i) {
                list.add(xi);
            }
        });
        return BoundSnapshot.create(list);
    }

    /**
     * Closure creating the linear of the xi variables
     * 
     * @return the linear
     */
    public Linear createLinear() {
        final Linear linear = pb.lp.createLinear();
        execute(new Closure() {
            @Override
            public void execute(Variable xi, int i) {
                if (xi == null) return;
                linear.add(pb.lp.createTerm(SolveRunnable.ONE, xi));
            }
        });
        return linear;
    }

    /**
     * Create a binary variable with the name xi[i].
     * 
     * @return the variable
     */
    public XiVariableSet createXiVariable() {
        return execute(new Closure() {
            @Override
            public void execute(Variable xi, int i) {
                if (xi == null) {
                    String name = String.format("xi_%d", i); //$NON-NLS-1$
                    pb.xi[i] = pb.lp.addBinaryVariable(name);
                }
            }
        });
    }

    /**
     * Execute the given closure for the selected variable.
     * 
     * @param c
     *            the closure to be executed
     * 
     * @return this object for method chaining
     */
    public XiVariableSet execute(Closure c) {
        // Loop on employees (i)
        for (int i = (iEq != null ? iEq : 0); i <= (iEq != null ? iEq : pb.getEmployees().size() - 1); i++) {
            if (matchesPredicates(pb.xi[i], i)) {
                c.execute(pb.xi[i], i);
            }
        }
        return this;
    }

    /**
     * Sets the i value representing the employee and modifying the execution of the closure
     * 
     * @param i
     *            index of the employee
     * @return this instance for method chaining.
     */
    public XiVariableSet iEq(int i) {
        this.iEq = i;
        return this;
    }

    /**
     * Closure checking if the vairable set is empty. Returns true if the variable set is empty
     * 
     * @return True if the variable set is empty. False if the variable set contains at least one variable.
     */
    public boolean isEmpty() {
        try {
            execute(new Closure() {
                @Override
                public void execute(Variable xi, int i) {
                    if (xi == null) return;
                    throw new NotEmptyException();
                }
            });
        } catch (NotEmptyException e) {
            // At least one variable is present.
            return false;
        }
        return true;
    }

    /**
     * Closure checking the bounds of the variables. Returns True if all the variables do not have the given bounds
     * value.
     * 
     * @param value
     *            the value of the bounds to check for
     * @return True if every variable bounds are sets to the given <code>value</code>. False if one variable bounds is
     *         not fixed or if the variable set is empty.
     */
    public boolean isFixedTo(final Double value) {
        Validate.notNull(value);
        try {
            execute(new Closure() {
                @Override
                public void execute(Variable xi, int i) {
                    if (xi == null) return;
                    if (!value.equals(xi.getLowerBound()) || !value.equals(xi.getUpperBound())) {
                        throw new NotFixedException();
                    }
                }
            });
        } catch (NotFixedException e) {
            // At least one variable is not fixed to the given value.
            return false;
        }
        return true;
    }

    /**
     * Check if the predicates affect the given variable.
     * 
     * @param xi
     *            The variable to check
     * @param i
     *            The index of the employee in the current variable
     * @return true if none of the predicates affect the given variable
     */
    private boolean matchesPredicates(Variable xi, int i) {
        if (this.predicates != null) {
            for (Predicate p : this.predicates) {
                if (!p.evaluate(xi, i)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Creates a new instance of the predicate for XiVariableSet if none is present with the given predicate or adds the
     * given predicate to the current one
     * 
     * @param p
     *            The predicate to add for this XiVariableSet
     * 
     * @return this instance for method chaining.
     */
    public XiVariableSet predicate(Predicate p) {
        if (this.predicates == null) {
            this.predicates = new ArrayList<XiVariableSet.Predicate>();
        }
        this.predicates.add(p);
        return this;
    }

    /**
     * Sets the lower and upper bound of the variable xi
     * 
     * @param lower
     *            the lower bound
     * @param upper
     *            the upper bound
     * @return this class for method chaining.
     */
    public XiVariableSet setBounds(final Number lower, final Number upper) {
        return execute(new Closure() {
            @Override
            public void execute(Variable xi, int i) {
                if (xi == null) return;
                xi.setLowerBound(lower);
                xi.setUpperBound(upper);

            }
        });
    }

    /**
     * Returns the size of the instance of XiVar
     * 
     * @return count, the size of the XiVariable
     */
    public int size() {
        final MutableInt count = new MutableInt(0);
        execute(new Closure() {
            @Override
            public void execute(Variable xi, int i) {
                count.increment();
            }
        });
        return count.intValue();
    }

    /**
     * Closure checking the value of the variables. Return false if 1 of the variable does not have the given value
     * 
     * @param value
     *            the value to check for
     * @return true if all the variables have the given value
     */
    public Boolean valueEq(final Double value) {
        Validate.notNull(value);
        if (isEmpty()) return false;
        try {
            // For cases where there are no variables
            execute(new Closure() {
                @Override
                public void execute(Variable xi, int i) {
                    if (xi == null) return;
                    if (!value.equals(xi.getValue())) {
                        throw new ValueNotEqException();
                    }
                }
            });
            return true;
        } catch (ValueNotEqException e) {
            return false;
        }
    }
}
