/**
 * Copyright(C) 2013 Patrik Dufresne Service Logiciel <info@patrikdufresne.com>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package com.planimod.core.planif;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.Validate;

import com.patrikdufresne.ilp.Linear;
import com.patrikdufresne.ilp.Variable;
import com.patrikdufresne.ilp.util.BoundSnapshot;
import com.planimod.core.Team;

/**
 * Gives a simple interface to manipulate the xij variables
 * <p>
 * Method chaining will filter the variables to construct a readable expression. Can also use operations on each
 * variables on selected group of variables.
 * 
 * @param pb
 *            the problemBuilder containing the array of variables to manipulate.
 * @author Karl
 * 
 */
public class XijVariableSet {

    /**
     * Close you execute for each variable.
     * 
     * @author Patrik Dufresne
     * 
     */
    public static interface Closure {
        public void execute(Variable xij, int i, int j);
    }

    /**
     * Trigger when a variable is not Empty.
     * 
     * @see XijVariableSet#isEmpty()
     * @author Karl
     */
    private static class NotEmptyException extends RuntimeException {
        // Nothing to do
    }

    /**
     * Predicate to trim the variables
     * 
     * @author Patrik Dufresne
     * 
     */
    public static interface Predicate {
        public boolean evaluate(Variable xij, int i, int j);
    }

    /**
     * Value of i. The index of the employee
     */
    private Integer iEq;

    /**
     * Value of j. The index of the team
     */
    private Integer jEq;

    /**
     * The problemBuilder associated
     */
    private ProblemBuilder pb;

    /**
     * Predicate list.
     */
    private List<Predicate> predicates;

    /**
     * Gives a simple interface to manipulate the xij variables
     * <p>
     * Method chaining will filter the variables to construct a readable expression. Can also use operations on each
     * variables on selected group of variables.
     * 
     * @param pb
     *            the problemBuilder containing the array of variables to manipulate.
     */
    protected XijVariableSet(ProblemBuilder pb) {
        Validate.notNull(this.pb = pb);
    }

    /**
     * Closure creating the linear of the xij variables
     * 
     * @return the linear
     */
    public Linear createLinear() {
        final Linear linear = pb.lp.createLinear();
        execute(new Closure() {
            @Override
            public void execute(Variable xij, int i, int j) {
                if (xij == null) return;
                linear.add(pb.lp.createTerm(SolveRunnable.ONE, xij));
            }
        });
        return linear;
    }

    /**
     * Create a binary variable with the name xij[i, j].
     * 
     * @return the variable
     */
    public XijVariableSet createXijVariable() {
        return execute(new Closure() {
            @Override
            public void execute(Variable xij, int i, int j) {
                if (xij == null) {
                    String name = String.format("xij_%d_%d", i, j); //$NON-NLS-1$
                    pb.xij[i][j] = pb.lp.addBinaryVariable(name);
                }
            }
        });
    }

    /**
     * Execute the given closure for the selected variable.
     * 
     * @param c
     *            the closure to be executed
     * 
     * @return this object for method chaining
     */
    public XijVariableSet execute(Closure c) {
        // Loop on employees (i)
        for (int i = (iEq != null ? iEq : 0); i <= (iEq != null ? iEq : pb.xi.length - 1); i++) {
            // Loop on teams (j)
            for (int j = (jEq != null ? jEq : 0); j <= (jEq != null ? jEq : pb.xij[i].length - 1); j++) {
                if (matchesPredicates(pb.xij[i][j], i, j)) {
                    c.execute(pb.xij[i][j], i, j);
                }
            }
        }
        return this;
    }

    /**
     * Sets the i value representing the employee and modifying the execution of the closure
     * 
     * @param i
     *            index of the employee
     * @return this instance for method chaining.
     */
    public XijVariableSet iEq(int i) {
        this.iEq = i;
        return this;
    }

    /**
     * Closure checking if the vairable set is empty. Returns true if the variable set is empty
     * 
     * @return True if the variable set is empty. False if the variable set contains at least one variable.
     */
    public boolean isEmpty() {
        try {
            execute(new Closure() {
                @Override
                public void execute(Variable xij, int i, int j) {
                    if(xij==null) return;
                    throw new NotEmptyException();
                }
            });
        } catch (NotEmptyException e) {
            // At least one variable is present.
            return false;
        }
        return true;
    }

    /**
     * Sets the j value representing the team index and modifying the execution of the closure
     * 
     * @param j
     *            index of the team
     * @return this instance for method chaining.
     */
    public XijVariableSet jEq(int j) {
        this.jEq = j;
        return this;
    }

    /**
     * Sets the team index value and modifies the execution of the closure
     * 
     * @param team
     *            the team
     * @return this instance for method chaining.
     */
    public XijVariableSet jEq(Team team) {
        Validate.notNull(team);
        return jEq(pb.getTeams().indexOf(team));
    }

    /**
     * Adds to the predicate to evaluate if the team is not the given team
     * 
     * @param j
     *            index of the team
     * @return this instance for method chaining.
     */
    public XijVariableSet jNotEq(final Team team) {
        Validate.notNull(team);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xij, int i, int j) {
                return !team.equals(pb.getTeams().get(j));
            }
        });
    }

    /**
     * Closure creating a bound snapshot
     * 
     * @return the snapshot
     */
    public BoundSnapshot createBoundSnapshot() {
        final List<Variable> list = new ArrayList<Variable>();
        execute(new Closure() {
            @Override
            public void execute(Variable xij, int i, int j) {
                if (xij != null) {
                    list.add(xij);
                }
            }
        });
        return BoundSnapshot.create(list);
    }

    /**
     * Check if the predicates affect the given variable.
     * 
     * @param xij
     *            The variable to check
     * @param i
     *            The index of the employee in the current variable
     * @param j
     *            The index of the team in the current variable
     * @return true if none of the predicates affect the given variable
     */
    private boolean matchesPredicates(Variable xij, int i, int j) {
        if (this.predicates != null) {
            for (Predicate p : this.predicates) {
                if (!p.evaluate(xij, i, j)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Check if the variable bounds are already set.
     * 
     * @return this instance for method chaining.
     */
    public XijVariableSet notFixed() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xij, int i, int j) {
                if (xij == null) return false;
                return !xij.getLowerBound().equals(xij.getUpperBound());
            }
        });
    }

    /**
     * Creates a new instance of the predicate for XijVariableSet if none is present with the given predicate or adds
     * the given predicate to the current one
     * 
     * @param p
     *            The predicate to add for this XijVariableSet
     * 
     * @return this instance for method chaining.
     */
    public XijVariableSet predicate(Predicate p) {
        if (this.predicates == null) {
            this.predicates = new ArrayList<XijVariableSet.Predicate>();
        }
        this.predicates.add(p);
        return this;
    }

    /**
     * Closure setting the lower and upper bound of the variables to the given value.
     * 
     * @param lower
     *            the lower bound
     * @param upper
     *            the upper bound
     * @return this class for method chaining.
     */
    public XijVariableSet setBounds(final Number lower, final Number upper) {
        return execute(new Closure() {
            @Override
            public void execute(Variable xij, int i, int j) {
                if (xij == null) return;
                xij.setLowerBound(lower);
                xij.setUpperBound(upper);
            }
        });
    }
}
