/**
 * Copyright(C) 2013 Patrik Dufresne Service Logiciel <info@patrikdufresne.com>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package com.planimod.core.planif;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.Validate;
import org.apache.commons.lang3.mutable.MutableInt;

import com.patrikdufresne.ilp.Linear;
import com.patrikdufresne.ilp.Variable;
import com.patrikdufresne.ilp.util.BoundSnapshot;
import com.planimod.core.Employee;
import com.planimod.core.Position;
import com.planimod.core.Section;
import com.planimod.core.Shift;
import com.planimod.core.Task;
import com.planimod.core.Team;
import com.planimod.core.TimeRanges;

/**
 * Gives a simple interface to manipulate the xijk variables
 * <p>
 * Method chaining will filter the variables to construct a readable expression. Can also use operations on each
 * variables on selected group of variables.
 * 
 * @param pb
 *            the problemBuilder containing the array of variables to manipulate.
 * @author Karl
 * 
 */
public class XijkVariableSet {

    /**
     * Close you execute for each variable.
     * 
     * @author Patrik Dufresne
     * 
     */
    public static interface Closure {
        public void execute(Variable xijk, int i, int j, int k);
    }

    /**
     * Trigger when a variable is not Empty.
     * 
     * @see XijkVariableSet#isEmpty()
     * @author Karl
     */
    private static class NotEmptyException extends RuntimeException {
        // Nothing to do
    }

    /**
     * Predicate to trim the variables
     * 
     * @author Patrik Dufresne
     * 
     */
    public static interface Predicate {
        public boolean evaluate(Variable xijk, int i, int j, int k);
    }

    /**
     * Trigger when a variable's value is diffrent from the given value.
     * 
     * @see XijkVariableSet#isValueEq()
     * @author Karl
     */
    private static class ValueNotEqException extends RuntimeException {
        // Nothing to do
    }

    /**
     * Value of i. The index of the employee.
     */
    private Integer iEq;

    /**
     * Value of j. The index of the team.
     */
    private Integer jEq;

    /**
     * Value of k. The index of the task.
     */
    private Integer kEq;

    /**
     * The problemBuilder associated.
     */
    private ProblemBuilder pb;

    /**
     * Predicate list.
     */
    private List<Predicate> predicates;

    /**
     * Gives a simple interface to manipulate the xijk variables
     * <p>
     * Method chaining will filter the variables to construct a readable expression. Can also use operations on each
     * variables on selected group of variables.
     * 
     * @param pb
     *            the problemBuilder containing the array of variables to manipulate.
     */
    protected XijkVariableSet(ProblemBuilder pb) {
        Validate.notNull(this.pb = pb);
    }

    /**
     * Check if the task (k) is classified
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet classified() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return pb.getTask(k).getPosition().getClassified();
            }
        });
    }

    /**
     * Closure creating a bound snapshot.
     * 
     * @return the snapshot
     */
    public BoundSnapshot createBoundSnapshot() {
        final List<Variable> list = new ArrayList<Variable>();
        execute(new Closure() {
            @Override
            public void execute(Variable xijk, int i, int j, int k) {
                if (xijk != null) {
                    list.add(xijk);
                }
            }
        });
        return BoundSnapshot.create(list);
    }

    /**
     * Closure creating the linear of the xijk variables
     * 
     * @return the linear
     */
    public Linear createLinear() {
        final Linear linear = pb.lp.createLinear();
        execute(new Closure() {
            @Override
            public void execute(Variable xijk, int i, int j, int k) {
                if (xijk == null) return;
                linear.add(pb.lp.createTerm(SolveRunnable.ONE, xijk));
            }
        });
        return linear;
    }

    /**
     * Closure creating the variable for the linear problem
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet createXijkVariable() {
        return execute(new Closure() {
            @Override
            public void execute(Variable xijk, int i, int j, int k) {
                if (xijk == null) {
                    String name = String.format("xijk_%d_%d_%d", i, j, k); //$NON-NLS-1$
                    pb.xijk[i][j][k] = pb.lp.addBinaryVariable(name);
                }
            }
        });
    }

    /**
     * Execute the given closure for the selected variable.
     * 
     * @param c
     *            the closure to be executed
     * @return this object for method chaining
     */
    public XijkVariableSet execute(Closure c) {
        // Loop on employees (i)
        for (int i = (iEq != null ? iEq : 0); i <= (iEq != null ? iEq : pb.xijk.length - 1); i++) {
            // Loop on teams (j)
            for (int j = (jEq != null ? jEq : 0); j <= (jEq != null ? jEq : pb.xijk[i].length - 1); j++) {
                // Loop on tasks (k)
                for (int k = (kEq != null ? kEq : 0); k <= (kEq != null ? kEq : pb.xijk[i][j].length - 1); k++) {
                    if (matchesPredicates(pb.xijk[i][j][k], i, j, k)) {
                        c.execute(pb.xijk[i][j][k], i, j, k);
                    }
                }
            }
        }
        return this;
    }

    /**
     * Sets the employee value and modifying the execution of the closure
     * 
     * @param emp
     *            the employee
     * @return this instance for method chaining.
     */
    public XijkVariableSet iEq(Employee emp) {
        Validate.notNull(emp);
        return iEq(pb.getEmployees().indexOf(emp));
    }

    /**
     * Sets the i value representing the employee and modifying the execution of the closure
     * 
     * @param i
     *            index of the employee
     * @return this instance for method chaining.
     */
    public XijkVariableSet iEq(int i) {
        this.iEq = i;
        return this;
    }

    /**
     * Closure checking if the variable set is empty. Returns true if the variable set is empty
     * 
     * @return True if the variable set is empty. False if the variable set contains at least one variable.
     */
    public boolean isEmpty() {
        try {
            execute(new Closure() {
                @Override
                public void execute(Variable xijk, int i, int j, int k) {
                    if (xijk == null) return;
                    throw new NotEmptyException();
                }
            });
        } catch (NotEmptyException e) {
            // At least one variable is present.
            return false;
        }
        return true;
    }

    /**
     * Check if the task (k) is in the real team (RTj)
     * 
     * @param j
     *            the index of the team with who to test
     * @param k
     *            the index of the task to be tested
     * @return this instance for method chaining.
     */
    private boolean iskInRTj(int j, int k) {
        return pb.getTeams().get(j).equals(pb.getTask(k).getProductionEvent().getShift().getTeam());
    }

    /**
     * Closure checking the value of the variables. Return true if all the variable do not have the given value
     * 
     * @param value
     *            the value to check for
     * @return True if all the variables have the given value. False if at least one of the value is different than the
     *         given <code>value</code> or if the variable set is empty.
     */
    public Boolean isValueEq(final Double value) {
        Validate.notNull(value);
        if (isEmpty()) return false;
        try {
            execute(new Closure() {
                @Override
                public void execute(Variable xijk, int i, int j, int k) {
                    if (xijk == null) return;
                    if (!value.equals(xijk.getValue())) {
                        throw new ValueNotEqException();
                    }
                }
            });
        } catch (ValueNotEqException e) {
            return false;
        }
        return true;
    }

    /**
     * Sets the j value representing the team and modifying the execution of the closure with the index <code>j</code>
     * 
     * @param j
     *            index of the team
     * @return this instance for method chaining.
     */
    public XijkVariableSet jEq(int j) {
        this.jEq = j;
        return this;
    }

    /**
     * Sets the j value representing the team and modifying the execution of the closure with the team <code>team</code>
     * 
     * @param j
     *            index of the team
     * @return this instance for method chaining.
     */
    public XijkVariableSet jEq(Team team) {
        Validate.notNull(team);
        this.jEq = pb.getTeams().indexOf(team);
        return this;
    }

    /**
     * Check if the team is not the given team
     * 
     * @param j
     *            index of the team
     * @return this instance for method chaining.
     */
    public XijkVariableSet jNotEq(final Team team) {
        Validate.notNull(team);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return !team.equals(pb.getTeams().get(j));
            }
        });
    }

    /**
     * Sets the j value representing the team and modifying the execution of the closure
     * 
     * @param j
     *            index of the team
     * @return this instance for method chaining.
     */
    public XijkVariableSet kEq(int k) {
        this.kEq = k;
        return this;
    }

    /**
     * Check if task (k) is part of the given position (pos)
     * 
     * @param pos
     *            the position to test
     * @return this instance for method chaining.
     */
    public XijkVariableSet kInPos(final Position pos) {
        Validate.notNull(pos);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return pos.equals(pb.getTask(k).getPosition());
            }
        });
    }

    /**
     * Check if the task (k) is part or not of real team (RTj).
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet kInRTj() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return iskInRTj(j, k);
            }
        });
    }

    /**
     * Check if the task (k) is part of real team (RTj).
     * 
     * @param j
     *            The index of the team to test
     * @return this instance for method chaining.
     */
    public XijkVariableSet kInRTj(final int j) {
        Validate.notNull(j);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j2, int k) {
                return iskInRTj(j, k);
            }

        });
    }

    /**
     * Check if the task (k) is in the real team (j)
     * 
     * @param team
     *            the team to test
     * @return this instance for method chaining.
     */
    public XijkVariableSet kInRTj(Team team) {
        Validate.notNull(team);
        return kInRTj(pb.getTeams().indexOf(team));
    }

    /**
     * Check if the task (k) is part of the given section (sec)
     * 
     * @param sec
     *            the index of the section
     * @return this instance for method chaining.
     */
    public XijkVariableSet kInSection(final Section sec) {
        Validate.notNull(sec);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return sec.equals(pb.getTask(k).getPosition().getSection());
            }
        });
    }

    /**
     * Check if the task (k) is part of the given shift (s1)
     * 
     * @param s1
     *            the shift with who to test
     * @return this instance for method chaining.
     */
    public XijkVariableSet kInShift(final Shift s1) {
        Validate.notNull(s1);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                Shift s2 = pb.getTask(k).getProductionEvent().getShift();
                return (s1.getTeam().equals(pb.getTeams().get(j))) && (s2.equals(s1) || TimeRanges.containsEquals(s2, s1) || TimeRanges.containsEquals(s1, s2));
            }
        });
    }

    /**
     * Check if the task (k) is part of the team (j)
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet kInTj() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                Task taskk = pb.getTask(k);
                Shift shiftOfk = taskk.getProductionEvent().getShift();
                if (shiftOfk.getTeam().equals(pb.getTeams().get(j))) {
                    return true;
                }
                if (taskk.getPosition().getSwappable()) {
                    for (Shift shift : pb.getShifts()) {
                        if (shift.getTeam().equals(pb.getTeams().get(j))
                                && (TimeRanges.containsEquals(shift, shiftOfk) || TimeRanges.containsEquals(shiftOfk, shift))) {
                            return true;
                        }
                    }
                }
                return false;
            }
        });
    }

    /**
     * Check if the task (k) is part or not of real team (RTj).
     * 
     * @param inRTj
     *            True if k must be in RTj and false if k must not be in Rtj
     * @return this instance for method chaining.
     */
    public XijkVariableSet kNotInRTj() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return !iskInRTj(j, k);
            }
        });
    }

    /**
     * Check if the predicates affect the given variable.
     * 
     * @param xijk
     *            The variable to check
     * @param i
     *            The index of the employee in the current variable
     * @param j
     *            The index of the team in the current variable
     * @param k
     *            The index of the task in the current variable
     * @return true if none of the predicates affect the given variable
     */
    private boolean matchesPredicates(Variable xijk, int i, int j, int k) {
        if (this.predicates != null) {
            for (Predicate p : this.predicates) {
                if (!p.evaluate(xijk, i, j, k)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Check if the employees are available to work on tasks
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet notAvailable() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return !pb.isAvailable(pb.getEmployees().get(i), pb.getTask(k));
            }
        });
    }

    /**
     * Check if the variable bounds are not already set to the same value
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet notFixed() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                if (xijk == null) return false;
                return !xijk.getLowerBound().equals(xijk.getUpperBound());
            }
        });
    }

    /**
     * Check if the variable bounds are already set to the same given value
     * 
     * @param value
     *            the Double to compare with the bounds value
     * @return this instance for method chaining.
     */
    public XijkVariableSet notFixedTo(final Double value) {
        Validate.notNull(value);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                if (xijk == null) return false;
                return !value.equals(xijk.getLowerBound()) || !value.equals(xijk.getUpperBound());
            }
        });
    }

    /**
     * Creates a new instance of the predicate for XijkVariableSet if none is present with the given predicate or adds
     * the given predicate to the current one
     * 
     * @param p
     *            The predicate to add for this XijkVariableSet
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet predicate(Predicate p) {
        if (this.predicates == null) {
            predicates = new ArrayList<XijkVariableSet.Predicate>();
        }
        this.predicates.add(p);
        return this;
    }

    /**
     * Check if the employee is qualified for the tasks k
     * 
     * @return this instance for method chaining.
     */
    public XijkVariableSet qualified() {
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                return pb.qualificationTable.containsEntry(pb.getEmployees().get(i), pb.getTask(k));
            }
        });
    }

    /**
     * Closure setting the lower and upper bound of the variable.s
     * 
     * @param lower
     *            the lower bound
     * @param upper
     *            the upper bound
     * @return this class for method chaining.
     */
    public XijkVariableSet setBounds(final Number lower, final Number upper) {
        return execute(new Closure() {
            @Override
            public void execute(Variable xijk, int i, int j, int k) {
                if (xijk == null) return;
                xijk.setLowerBound(lower);
                xijk.setUpperBound(upper);
            }
        });
    }

    /**
     * Returns the size of the instance of XijkVariable
     * 
     * @return count, the size of the XijkVariable
     */
    public int size() {
        final MutableInt count = new MutableInt(0);
        execute(new Closure() {
            @Override
            public void execute(Variable xijk, int i, int j, int k) {
                if (xijk == null) return;
                count.increment();
            }
        });
        return count.intValue();
    }

    /**
     * Check if the value of the variable is the given value
     * 
     * @param value
     *            the Double to compare with the bounds value
     * @return this instance for method chaining.
     */
    public XijkVariableSet valueEq(final Double value) {
        Validate.notNull(value);
        return predicate(new Predicate() {
            @Override
            public boolean evaluate(Variable xijk, int i, int j, int k) {
                if (xijk == null) return false;
                return value.equals(xijk.getValue());
            }
        });
    }
}
